/* ---------------------------------------------------------------------- *
 *                         inputs.c : May 1, 1993.
 * This file contains code to produce random inputs of the input_dimension
 * specified in the simulation file. Each component of the input_vector is
 * uniformly randomly distributed in the range specified in the simulation
 * file. Scaling functions are also provided here for the convenience of
 * display_net and initialization routines. The scaling functions ae used
 * only for displaying (x,y) co-ords of inputs and drawing weights.
 * ---------------------------------------------------------------------- */

#include "latdefs.h"

extern int input_dimension              ;
extern Ranges *input_ranges             ;
extern double *input_vectors, *raw_input;
extern int distribution                 ;/* The input distribution chosen */ 

extern void tocartesian()               ;/* In lcsofm.c: Normalizer       */
extern double drand48()                 ;

void uniform()   ;
void normal()    ;
void plus()      ; /* The size of the plus is hardcoded */
void narrowplus(); /* The size is even smaller. */

/* ---------------------------------------------------------------------- */

double scalelu(i,f01)   /* Scale component i between lower & upper limits */
int i; double f01;
{
  return (input_ranges[i].low + f01*(input_ranges[i].high -
				     input_ranges[i].low));
}
/* ---------------------------------------------------------------------- */
/* The scalelu and scale01 functions are inverses of each other 
 */
double scale01(i,f01)   /* Scale component i back to 0-1 range            */
int i; double f01;
{
  return ((f01-input_ranges[i].low)/(input_ranges[i].high -
				     input_ranges[i].low) );
}
/* ---------------------------------------------------------------------- */
void generate_input()
{
  /* Create input angles distributed within the given ranges
   * and convert to cartesian coordinates. For normalized input, the 
   * input dimension is input_dimension+1 
   */
  int i; double temp;
  double t[MAXDIM]  ;

  switch(distribution)
    {
    case 0: uniform();
      break;

    case 1: normal();
      break;

    case 2: plus(); break;

    case 3: narrowplus(); break;
    }
  /* Save the raw input in vector raw_input */
  for(i=0; i< input_dimension; i++)  raw_input[i]= input_vectors[i];

  tocartesian(input_vectors,t); /* Convert to cartesian co-ords */
 
  for(i=0; i<=input_dimension; i++) input_vectors[i] = t[i];
}
/* ---------------------------------------------------------------------- */
void uniform()
{
  int i;

  for(i=0; i<input_dimension; i++)
    input_vectors[i] = scalelu(i,drand48());
}
/* ---------------------------------------------------------------------- */
void normal()
{
  uniform();
}
/* ---------------------------------------------------------------------- */
void plus()
{
#define LOWER 0.35
#define UPPER 0.65

  extern double drand48();   double x,y;

  for(x = drand48(), y = drand48(); 
      ((x < LOWER) || (x > UPPER)) && ((y < LOWER) || (y > UPPER));
      x = drand48(), y = drand48());
      
  input_vectors[0] = scalelu(0,x);
  input_vectors[1] = scalelu(1,y);
}
/* ---------------------------------------------------------------------- */
void narrowplus()
{
#define LOWER 0.45
#define UPPER 0.55

  extern double drand48();   double x,y;

  for(x = drand48(), y = drand48(); 
      ((x < LOWER) || (x > UPPER)) && ((y < LOWER) || (y > UPPER));
      x = drand48(), y = drand48());
        
  input_vectors[0] = scalelu(0,x);
  input_vectors[1] = scalelu(1,y);
}
/* ---------------------------------------------------------------------- */
