/*                                                                           *
 * --------------------- File : latdefs.h ------------------------------- *
 * Created on April 17,93 for lcsofm.c
 * Header file for a self organizing feature map with one layer. The lateral *
 * weights can modify themselves according to activity. The lateral          *
 * excitation, inhibition and input connections are normalized separately.   *
 * Lateral inhibition/excitation in the maps concentrate the activity.       *
 * The display uses motif and X windows.                                     *
 * Each layer is an NMAX by NMAX square grid of neurons.                     *
 *                                                                           *
 * ------------------------------------------------------------------------- */



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifdef CRAY
#include <malloc.h>
#endif


#ifdef XWINDOW                       /* Include these only if X is defined.  */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/Command.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#endif

#define True            1
#define False           0
#define CHAR            0
#define INT             1
#define FLOAT           2
#define ROUND(a)        (floor( (0.5 + (a))))

#define MAXDIM          100        /* Maximum input dimension               */
#define SNAPSHOTEND     999999     /* Marker for last snapshot              */
#define MAXSNAPS        100
#define MAXPHASES       100

#define DEATH_FLAG      0.00000001 /* Marker for dead connection: -ve for exc 
				    * wts and +ve for inh wts. The value is 
				    * small, to save checks in lat_resp()
				    */

#define MAXKILLPHASES   50       /* No. of connection death phases         */

#define INPUT    1
#define NET      2
#define NET_ACT  3
#define ATT_ACT  4

#define NMAX               40    /* Maximum size of grid                   */

 
#define COLOR_RANGE        234   /* Colors from 0...234 - 235 colors       */
                                 /* This one should be divisible by 3.     */

/* ---------------------- Neuron definitions. ---------------------------- */ 
                          
typedef struct N                 /* The record type for neurons in a map   */
{
  double beta,delta   ;          /* Sigmoid upper and lower thresholds     */
  double *weights     ;          /* Weights - these are a malloced array   */
  double *lat_exc_wts ;          /* Lat exc weights are stored column by   */
  double *lat_inh_wts ;          /* column.                                */
} Neuron              ;

typedef struct r
{
  double low, high    ;          /* Range for each input vector component  */
  double randomness   ;          /* Initial randomness of each component   */
}Ranges               ;

typedef struct board             /* Communication board for User Interface */
{ 
  char param_name[40];           /* Name of parameter.                     */
  int data_type      ;           /* Param data_type:CHAR=0,INT=1,FLOAT=2   */
  int update_flag    ;           /* If set, the data must be updated by UI */
  union t{
    char   *c        ;           /* This union holds the addresses of the..*/
    int    *i        ;           /* associated data item. The locations .. */
    double *f        ;           /* they point to can be changed by the UI */
  }data_ptr          ;
} ParamInterface     ;

typedef struct b{                /* The parameter blackboard               */
  int no_of_entries      ;
  ParamInterface list[50];
}Blackboard              ;
/* -------------------- Display definitions. ----------------------------- */
#ifdef XWINDOW  

#define XtPointer caddr_t        /* Compatibility's sake                   */
#define XmStringCreateSimple(a) XmStringLtoRCreate((a),XmSTRING_DEFAULT_CHARSET)

typedef struct inpt              /* To store previous input vectors.       */
{
  double *input_vector;
  struct inpt *next   ;
}Input_history        ;

typedef struct info              /* Display window info                    */
{
  Display *d ;
  Window   w ;
  GC      gc ;
  Pixel fg,bg;
  int     width, height; 
}d_info;

typedef struct d                 /* The current display configuration     */ 
{
  d_info  input,                 /* Input window -random dots in 2d space */
          net,                   /* 2d net display                        */
          net_act,               /* net activity pattern display          */
          att_act;               /* upper layer activity display          */
  XColor  colors[255];           /* Colormap of the application. Max 255  */
  Widget  stop_button,           /* Other widgets the application needs...*/
      continue_button,           /* to know about. Mostly to set the...   */
      restart_button ,           /* sensitivity.                          */
      file_button    ,
      command_w      ,           /* For filenames & changing parameters   */
      list_w         ;           /* For displaying parameters & values    */
} My_Display         ;

#endif
/* ----------------------------------------------------------------------- */
