/* ---------------------------------------------------------------------- 
 *
 * Revised version of LCSOFM. April 17, 1993.
 * Written for the CRAY-YMP, but can run on all systems.
 * 
 * Features:
 * 1. Real sigmoid.
 * 2. Standard user interface with world.
 * ---------------------------------------------------------------------- */


#include "globals.h"

extern void srand48()              ;
extern double drand48()            ;
extern void display_init()         ; 
extern void generate_input()       ;
extern double scalelu()            ;
extern size_t fread()              ;

#ifdef XWINDOW
extern void display_activity()     ; /* In xdraw.c  */
extern void display_net()          ;
extern void display_dimension()    ;
extern void display_input()        ;
extern void display_parameters()   ; /* These are in xdisplay.c */
extern void display_message()      ;
extern void display_waiting_cursor()  ; 
extern void display_normal_cursor()   ;
extern void update_param_board()   ;
extern void wait_for_buttonpress() ; 
#endif

int  init_params()                 ;
int  iterate_snapshots()           ;
void training()                    ;
void present_inputs()              ;
void kill_connections()            ;
void settle_responses_and_display();
void compute_responses()           ;
double lat_resp()                  ;
void modify_weights()              ;
void modify_lat_wts()              ;
void modify_sigmoid()              ;
double sigmoid()                   ;
void init_weights()                ;
void read_params()                 ;
double distr_weight()              ;
void tocartesian()                 ;
void setup_latw_symmetric()        ;
void setup_latw_unsymmetric()      ;
double dist_lat_wt()               ;
int  save_current_binary()         ;
void weights_IO()                  ;
void iterate_weights()             ;
void writefun()                    ;
void readfun(), randfun()          ;
void setup_blackboard()            ;
void print_params()                ;

#ifdef XWINDOW
int  Xtraining()                   ;
void add2history()                 ;
void free_inptlist()               ;
#endif

char msg[100]                      ; /* Buffer for messages to display */
/* ---------------------------------------------------------------------- */

void main(argc, argv)
int argc;
char **argv;
{
  setup_blackboard(&blackboard); /* Setup the interface structure  */

#ifdef XWINDOW
  if (displaying) 
    {     /* Change the argc and argv later. */
      display_init(1, argv);    /* Set up all the widgets and menus    */
      XtMainLoop ()        ;    /* Start the main event-handling loop. */
      exit(0)              ;    /* Exit at the end of the application.
				 * Control never goes beyond this line if 
				 * displaying.
				 */
    }
#endif

/* The following functions are for doing the computation without a display.
 * The calls to the display in these functions are included only if XWINDOW is 
 * defined.
 */   

  if (!init_params(argv[1])) exit(0)   ;    /* File not found */
  if (continuing){ 
      if (!iterate_snapshots()) exit(0);    /* File problem   */
    }
  else init_weights()                  ;
  training()                           ;
  fclose(runtimefile)                  ;
  exit(0)                              ;
} 


/* ---------------------------------------------------------------------- */

int iterate_snapshots()

/* This procedure gets called to iterate the saved snapshots. It runs
 * through all the snapshots and returns true. If XWINDOW is defined, it
 * also updates parameters. changes cursors and waits for mouse button
 * press to go to next snapshot.
 * Weight changes during presentations are overwritten when the next snapshot
 * is retrieved. Only for the last snapshot is the weight change effective.
 * The program continues from there if there are more iterations to go.
 */
{
  int i, oldt  ;
  char buf[50] ;


  if ((fp=fopen(simufile,"r"))==NULL){
    fprintf(stderr,"Simufile %s couldn't be opened\n", simufile);
    return(0)                                                   ;
  }
  read_params(fp)          ;           /* Read from parameter file. */
  fclose(fp);

  sprintf(buf,"%s.bin",simufile);      /* Open the weights file */
  if ((fp=fopen(buf,"rb"))==NULL){
    fprintf(stderr,"Wt file %s couldn't be opened\n", buf);
    return(0)                                                   ;
  }
  
  oldt=0                  ;
  iterate_weights(randfun);            /* just to update rand                */
  while (fscanf(fp,"%d",&t)!=EOF)      /* read the t of the snapshot         */
    {
#ifdef XWINDOW
      display_waiting_cursor(&disp);   /* Waiting cursor while loading       */
      display_parameters(&disp)    ;   /* Update the parameter display       */
#endif
      startt=t ;
      for(i=0; i<startt-oldt; i++)     /* Generate one less input point */
	{ 
	  generate_input();
#ifdef XWINDOW
	  add2history(raw_input);/* Add points to the ptlist */
#endif
	}
      oldt=startt          ;
      weights_IO(fread,fp) ;  /* Read the current network from file          */
      generate_input();       /* Create the last input.                      */
#ifdef XWINDOW
      add2history(raw_input);/* Add last point               */
#endif
      present_inputs()     ; 

#ifdef XWINDOW 
      if (displaying) {  /* Here we have to sense a mouse event by ourselves */
	sprintf(buf,"Snapshot %d loaded. Press mouse button to continue.\n",t);
	display_message(buf)        ;  /* Display the message                */
	display_normal_cursor(&disp);  /* Revert back to normal cursor       */
	wait_for_buttonpress(&disp) ;  /* wait until mouse button is pressed */
      }
#endif
    }
  fclose(fp);

  /* Goto the next snapshot. */
  for (nextsnapshot=0; t >= snapshots[nextsnapshot]; nextsnapshot++);
  /* Goto next kill phase  */
  for (killphase=0; t >= kill_phases[killphase]; killphase++);
  return(1);
}  
/* ---------------------------------------------------------------------- */

void training()  
/* If X is being used, then this procedure should be substituted by a work
 * proc so that other events can be processed in the middle of the long
 * computation. So if we are using X, we bypass this procedure altogether,
 * and instead use a procedure called Xtraining() which does the same job,
 * but is more fancy.
 */
{

  for(t=startt; t<=tend; t++)
    {
      generate_input()                    ;
      if (t >= snapshots[nextsnapshot])     /* Save current network before   */
	save_current_binary()             ; /* modifying. Retrieve opposite. */

      if (t>=kill_phases[killphase]){
	kill_connections()                ;
	killphase++                       ;
      }
      present_inputs()                    ;
    }
}
/* ---------------------------------------------------------------------- */
#ifdef XWINDOW
int Xtraining(c_data)  
XtPointer c_data;
/* This work proc does the computation if there are no events to be processed.
 * For a work procedure to be called repeatedly, you must return false all the
 * time. 
 */
{
  char msg[100];
  int i;

  XmUpdateDisplay(top_level);
  if (Training_Start && !Training_Done)
    {
      if (t > tend){
	Training_Done=True                ;  /* End of training */

	XtSetSensitive(disp.stop_button    , False);
	XtSetSensitive(disp.continue_button, False);
	XtSetSensitive(disp.restart_button , True );
	XtSetSensitive(disp.command_w      , True );
	XtSetSensitive(disp.list_w         , True );
	XFlush(XtDisplay(top_level)               );

	sprintf(msg,"Training complete. Last Iteration = %d", t-1);
	display_message(msg);
	print_params(runtimefile,&blackboard); 
	/* Save final parameters of simulation from blackboard. */
      }

      for (i=0; i<disp_rate; i++, t++){

	generate_input()                    ;
	add2history(raw_input)              ;

	if (t >= snapshots[nextsnapshot])
	  save_current_binary()             ;
	if (t>=kill_phases[killphase]){
	  kill_connections()                ;
	  killphase++                       ; 
	}
	present_inputs()                    ;
      }
    }
  return(False);
}
#endif
/* ---------------------------------------------------------------------- */
void kill_connections()
{
/* Dead connections have a small negative value on them. These connections
 * are not modified afterwards. Nor do they make any significant contribution
 * to the response on the map because of their small value.
 */
  int i,j,k;
  
  for(i=0; i<N; i++)
    for(j=0;j<N; j++)
      {
	for(k=0; k<lat_exc_dimension; k++)
	  if (map[i][j].lat_exc_wts[k] < exc_death)
	    map[i][j].lat_exc_wts[k] = -DEATH_FLAG;

	for(k=0; k<lat_inh_dimension; k++)
	  if (map[i][j].lat_inh_wts[k] < inh_death)
	    map[i][j].lat_inh_wts[k] = -DEATH_FLAG;
      }
}
/* ------------------------------------------------------------------------ */
void present_inputs()
{
  int ui, uj;

#ifdef XWINDOW

  if (displaying && (t%disp_rate ==0)){  /* Display only this often         */
    display_parameters(&disp)         ;
    display_net(&disp, map, N, border);
/*
    display_dimension(&disp,map,N,border,2); */ /* Show third dimension        */

    display_input(&disp, input_list)  ;  
  }
#endif

  settle_responses_and_display();  /* Settle and show activity */
  modify_weights()              ;
  modify_lat_wts()              ;
  modify_sigmoid()              ;
}

/* ------------------------------------------------------------------------ *
 *                       Core of LCSOFM computations                        *
 * ------------------------------------------------------------------------ */

void init_border_act(border)/* Initialize border neurons to random activity */
int border;
{
  int i,j;

  for(i=0; i<N; ){                              
      if (i==border) i=N-border;
      for(j=0; j<N; j++ )
	map_activity[i][j] = resp_to_inp[i][j] = drand48();
      i++;
    }
  for(j=0; j<N; ){ 
    if (j==border) j=N-border;
    for(i=0; i<N; i++ )
      map_activity[i][j] = resp_to_inp[i][j] = drand48();
    j++;
  }
}
/* ------------------------------------------------------------------------ */
void settle_responses_and_display()
{
  int i,j,k,ts;
  double max_activity;

  if (border > 0) init_border_act(border); /* Initialize activity in border */

  for(i=border; i<N-border; i++)
    for(j=border; j<N-border; j++) /* Init rest of neurons, compute input act*/
      { 
	/* If dynamics=0, blank out previous activity in network.            */
	map_activity[i][j] = (dynamics==0)? 0.0 : map_activity[i][j];

	resp_to_inp[i][j] = 0.0;
	
	for(k=0; k<=input_dimension; k++) /* Afferent input activation.      */
	  resp_to_inp[i][j] += (input_vectors[k]) * (map[i][j].weights[k]);
      }


#ifdef XWINDOW
  /* Find the maximum of the activity for the input for comparison           */

  max_activity = 0.0; maxact_i=0; maxact_j=0;
  for(i=border; i<N-border; i++)
    for(j=border; j<N-border; j++)
      if (resp_to_inp[i][j] >= max_activity){
	max_activity = resp_to_inp[i][j];
	maxact_i=i; maxact_j=j;
      }
#endif

  /* Let the network settle into an activity bubble. */
  for(ts=1; ts<=tsettle; ts++)
    {
      compute_responses();  /* Add up input and lateral activations          */

#ifdef XWINDOW
      if (displaying && (t%disp_rate ==0))
	{
	  display_activity(&disp, map_activity, N, maxact_i, maxact_j);

	  sprintf(msg,"Settling iteration: %d",ts);
	  display_message(msg);

	  if (step_through) wait_for_buttonpress();
	}
#endif
    }
}

/* ---------------------------------------------------------------------- */
void compute_responses()
{
  register int i,j;

  /* Copy the the activity from the previous settling iteration.
   * This is used to compute the responses in the new iteration.
   */
  for(i=0; i<N; i++)   
    for(j=0; j<N; j++)
      prev_map_activity[i][j] = map_activity[i][j];

#ifdef CRAY
#pragma _CRI taskloop private(i,j) shared(N,map_activity,resp_to_inp,gammaexc,exc_rad,map,gammainh,inh_rad)
#endif
 
  /* Add up input, lateral excitation and lateral inhibition.
   */
  for(i=0; i<N; i++){
    for(j=0; j<N; j++){

      map_activity[i][j] = 
	sigmoid(resp_to_inp[i][j]                                   + 
		gammaexc*lat_resp(i, j, exc_rad, 
					     map[i][j].lat_exc_wts) -
		gammainh*lat_resp(i, j, inh_rad,
					     map[i][j].lat_inh_wts),
			  
		map[i][j].delta, map[i][j].beta);
    }
  }
}
/* ---------------------------------------------------------------------- */

double lat_resp(ui, uj, radius, weights)
register int ui, uj, radius;
register double *weights;
{
/* Computes lateral response excitation or inhibition with the given weight
 * vectors and radius. The weight vectors are normalized.
 * The computation loop is:
 *---------------------------------
 * for(i=lowi; i<=highi; i++)  
 *   for(j=lowj; j<=highj; j++)
 *	resp += (prev_map_activity[i][j] *
 *	 weights[(i-ui+radius) * array_width +(j-uj+radius)]);
 *---------------------------------
 * This loop is optimized below for scalar & vector machines separately....
 */
  register int i, j, lowi, highi, lowj, highj,count;
  register int array_width,optimize1,optimize2     ;
  register double resp=0.0                         ;
#ifdef CRAY
  register int optimize2_0,optimize2_1,optimize2_2,optimize2_3;
#endif

  array_width = 2*radius + 1               ;
  lowi =ui-radius ; if (lowi < 0) lowi =0  ;  /* The corners of lateral ...*/
  highi=ui+radius ; if (highi>=N) highi=N-1;  /* ... response              */
  lowj =uj-radius ; if (lowj < 0) lowj =0  ;
  highj=uj+radius ; if (highj>=N) highj=N-1;

  optimize1 = (-ui+radius)*array_width + (-uj+radius);

#ifndef CRAY
  for(i=lowi; i<=highi; i++){
    optimize2 = i*array_width + optimize1;
    for(j=lowj; j<=highj; j++)
	resp += (prev_map_activity[i][j] * weights[optimize2+j]);
  }
#else  /* CRAY is defined: unroll the loop 4 times for good vectorization   */
  for(i=lowi+3; i<=highi; i+=4){
    
    optimize2_0 = i*array_width + optimize1;
    optimize2_1 = (i-1)*array_width + optimize1;
    optimize2_2 = (i-2)*array_width + optimize1;
    optimize2_3 = (i-3)*array_width + optimize1;

    for(j=lowj; j<=highj; j++)
      resp += (prev_map_activity[i][j]   * weights[optimize2_0 + j]) +
	      (prev_map_activity[i-1][j] * weights[optimize2_1 + j]) +
	      (prev_map_activity[i-2][j] * weights[optimize2_2 + j]) +
	      (prev_map_activity[i-3][j] * weights[optimize2_3 + j]) ;
       		
  }
  if ((i-4) < highi){              /* Compute for the remaining values of i */
    for(i=i-3; i<=highi; i++){
      optimize2 = i*array_width + optimize1;
      for(j=lowj; j<=highj; j++)
	resp += (prev_map_activity[i][j] * weights[optimize2+j]);
    }
  }
#endif
  return(resp);
}
/* ------------------------------------------------------------------------ */
void modify_weights() /* Change afferent weights by normalized Hebbian rule */
{
  double length;
  int i,j,k   ;

  for(i=border; i<N-border; i++)
    for(j=border; j<N-border; j++)
      {
	length=0.0  ;
	for(k=0; k<=input_dimension; k++)
	  {
	    map[i][j].weights[k] += alpha_input * input_vectors[k] *
	                            map_activity[i][j];
	    length += map[i][j].weights[k] * map[i][j].weights[k];
	  }

	/* Normalize to keep length of wt. vector constant.*/
	length= 1.0/sqrt(length); 
	for(k=0; k<=input_dimension; k++)
	  map[i][j].weights[k] *= length;
      }  
}
/* ------------------------------------------------------------------------ */
void modify_lat_wts()
{
/* Modify lateral weights by normalized Hebbian rule. The normalization
 * keeps sum of lateral excitatory weights and sum of lateral inhibitory
 * weights constant and equal to 1.0.
 *
 * This function can be speeded up by the following:
 * 1. Symmetry: Weight change for neuron (i,j) and (k,l) are the same.
 * 2. Constant folding. alpha_exc*map_activity[i][j] is a loop constant.
 * 3. Vectorization: Eliminate the if statement in doing the multiplication
 *    and adding. Calculate the deltaws in another array, and update the
 *    connections for (i,j) and (k,l) checking for the DEATH_FLAG. Calculate
 *    total deltaw for normalization from the deltaw array. When updating
 *    a dead connection, subtract the deltaw for that one from the total.
 *    Use these two separate totals for normalization of (i,j) & (k,l).
 */
  int elowk, elowl, ehighk, ehighl,ilowk, ilowl, ihighk, ihighl;
  register int i,j,k,l,index,m,optimize1, optimize2,exc_ar_width, inh_ar_width;
  register double alpha, deltaW,new_length,*lat_exc_wts, *lat_inh_wts;
  register double const1;

  exc_ar_width = 2*exc_rad+1;
  inh_ar_width = 2*inh_rad+1;

#ifdef CRAY
#pragma _CRI taskloop private(i,j,k,l,deltaW,elowk, elowl, ehighk, ehighl,ilowk, ilowl, ihighk, ihighl,const1,optimize1,optimize2,new_length,index,m,lat_exc_wts,lat_inh_wts) shared(exc_ar_width,inh_ar_width,N,map_activity,exc_rad,map,inh_rad,alpha_exc,alpha_inh,lat_exc_dimension,lat_inh_dimension, symmetric_exc_wts)
#endif

  for(i=0; i<N; i++){ /* -------------------------------------------------- */

    if (symmetric_exc_wts){
      elowk  = i-exc_rad; if (elowk < 2*exc_rad) elowk = (i+1) >> 1;
      ehighk = i+exc_rad; if (ehighk>=N- 2*exc_rad) ehighk = N-1 -((N-i) >> 1);
    }
    else{
      elowk  = i-exc_rad; if (elowk < 0) elowk =0      ;
      ehighk = i+exc_rad; if (ehighk>=N) ehighk = N-1  ;
    }
    ilowk  = i-inh_rad; if (ilowk < 0) ilowk =0      ;
    ihighk = i+inh_rad; if (ihighk>=N) ihighk = N-1  ;
    for(j=0; j<N; j++){ /* ------------------------------------------------ */

      if (symmetric_exc_wts){
	elowl  = j-exc_rad; if (elowl < 2*exc_rad) elowl = (j+1) >> 1 ;
	ehighl = j+exc_rad; if (ehighl>=N-2*exc_rad) ehighl= N-1 -((N-j) >> 1);
      }
      else{
	elowl  = j-exc_rad; if (elowl < 0) elowl =0    ;
	ehighl = j+exc_rad; if (ehighl>=N) ehighl = N-1;
      }
      ilowl  = j-inh_rad; if (ilowl < 0) ilowl =0    ;
      ihighl = j+inh_rad; if (ihighl>=N) ihighl = N-1;

      lat_exc_wts = map[i][j].lat_exc_wts ;
      new_length=0.0;
      optimize1 = (-i+exc_rad)*exc_ar_width-j + exc_rad;
      const1 =  alpha_exc * map_activity[i][j];

      for(k=elowk; k<=ehighk; k++){

	optimize2 = k*exc_ar_width + optimize1;
	for(l=elowl; l<=ehighl; l++){

/* index = (k-i+exc_rad)*exc_ar_width + (l-j+exc_rad); */
	  deltaW = const1 * map_activity[k][l]; 
	  index = optimize2 + l; 
	  if (lat_exc_wts[index] >= 0.0){ /* Otherwise, dead connection      */
	    lat_exc_wts[index] += deltaW;
	    new_length += lat_exc_wts[index];
	  }   
	}
      }
#ifdef DEBUG3
      if (i==j) printf("exc: new_length = %lf\n",new_length);
#endif
      new_length = 1.0/new_length;
      for(m=0; m<lat_exc_dimension; m++){

	map[i][j].lat_exc_wts[m] *= new_length;
#ifdef DEBUG3
      if (i==j) printf("exc_wt[%d] = %0.6lf\n",m,map[i][j].lat_exc_wts[m]);
#endif
      }

      /* -------------------- Now modify the inhibitory weights -------------*/

      lat_inh_wts = map[i][j].lat_inh_wts;
      new_length=0.0;
      optimize1 = (-i+inh_rad)*inh_ar_width- j + inh_rad;
      for(k=ilowk; k<=ihighk; k++){

	optimize2 = k*inh_ar_width + optimize1;
	for(l=ilowl; l<=ihighl; l++){
	  
/* index = (k-i+inh_rad)*inh_ar_width + (l-j+inh_rad); */

	  index = optimize2 + l;
	  if (lat_inh_wts[index] >= 0.0){ /* Otherwise, dead connection      */
	    lat_inh_wts[index] += alpha_inh * map_activity[i][j] *
	                          map_activity[k][l];
	    new_length += lat_inh_wts[index];
	  }
	}
      }
#ifdef DEBUG3
      if (i==j) printf("inh: new_length = %lf\n",new_length);
#endif
      new_length = 1.0/new_length;

      for(m=0; m<lat_inh_dimension; m++){
	map[i][j].lat_inh_wts[m] *= new_length;
#ifdef DEBUG3
      if (i==j) printf("inh_wt[%d] = %0.6lf\n",m,map[i][j].lat_inh_wts[m]);
#endif
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
void modify_sigmoid()
{
/* The threshold mechanism is a combination of act_dependent & linear parts 
 */
/* The scaling factors for threshold modification are such that if the activity
 * were to be 1.0 for all the training iterations, the thresholds would reach
 * their limit values at tend from the starting values, provided alpha is 1.0
 * If alpha were > 1.0, they would reach the upper limit sooner. The scaling
 * factors are set in the read_params function.
 */ 
 int i,j;
 double rate_of_change1, rate_of_change2, const1, const2, const3, const4;

/* Change the threshold only if the final thresholds are not reached.
 */
 rate_of_change1 = alpha_delta * scale_delta    ;
 rate_of_change2 = alpha_beta * scale_beta      ;
 const1 = rate_of_change1 * act_dep_frac        ;
 const2 = rate_of_change1 * (1.0 - act_dep_frac);
 const3 = rate_of_change2 * act_dep_frac        ;
 const4 = rate_of_change2 * (1.0 - act_dep_frac);

 for(i=0; i<N; i++)
   for(j=0; j<N; j++){

     map[i][j].delta += ((map[i][j].delta < final_delta) ? 
			 (const1 *map_activity[i][j] + const2) : 0.0);
     
     map[i][j].beta -= ((map[i][j].beta > final_beta) ? 
			(const3 *map_activity[i][j] + const4) : 0.0);
   }			
}
/* ------------------------------------------------------------------------ */

double sigmoid(activity, delta, beta) /* Linear approx of sigmoid */
double activity; double delta, beta;  /* Lower & upper thresholds */    
{
  /* Transform the activity to a sigmoid response between 0 and 1.0.
   * Between the thresholds, response is linear.
   */
#ifdef DEBUG2
  printf("activity=%.6lf; delta=%.6lf, beta=%.6lf\n",activity,delta,beta);
#endif
  if (activity <  delta) return(0.0);

  else if (activity > beta) return(1.0);

  else return( (activity - delta) / (beta-delta)
	      );
}

/* ------------------------------------------------------------------------ *
 *                       Initializations                                    *
 * ------------------------------------------------------------------------ */
char* alloce(allocsize)/* Allocator with error check for mem unavailability */
int allocsize;
{
  char *new;

  if (( new = malloc(allocsize))==NULL){
    fprintf(stderr,"Not enough memory. Exiting\n");
    exit(0);
  }
  return(new);
}
/* ------------------------------------------------------------------------ */
int init_params(argv)
char *argv;
{
  int i, j, k, temp, allocsize;
  FILE *fp; char buf[50];

  sprintf(simufile, "%s", argv);
  if ( (fp=fopen(simufile,"r"))==NULL){
    fprintf(stderr, "File %s not found\n", simufile);
    return(0);
  }
  read_params(fp)  ;
  fclose(fp)       ; 
  print_params(stderr, &blackboard); /* Echo parameters to stderr */

  sprintf(buf,"%s.bin",simufile);
  if( ((fp=fopen(buf,"rb"))==NULL) || (fscanf(fp,"%d",&temp)==EOF) )
    continuing=0   ; /* File has no data */
  else continuing=1;
  fclose(fp)       ;
 
  sprintf(buf,"%s.run",simufile);
  if( (runtimefile=fopen(buf,"w"))==NULL)
    fprintf(stderr, "Runtime parameter file %s not opened\n", buf);
  else
    {      /* First fopen() clears the file. Next one opens for appending */
      fclose(runtimefile)         ;
      runtimefile = fopen(buf,"a");
    }
  lat_exc_dimension = (2*exc_rad + 1) * (2*exc_rad + 1);
  lat_inh_dimension = (2*inh_rad + 1) * (2*inh_rad + 1);

  /* Add a scaling factor for sigmoid change for convenience in setting up
   * the simulation files. This is set based on first value of "tend" in the
   * simulation file. Interactive changes to tend do not reflect on this value.
   * Initially, final_delta > delta, and final_beta < beta.
   */ 
  scale_delta = (final_delta - delta)/(double)tend;
  scale_beta  = (beta - final_beta)/ (double)tend ;

  for(i=0; i< N; i++)     /* Initialize sigmoids of all neurons, initialize */
    for(j=0; j< N; j++){  /* activity in the map.                           */
	map[i][j].delta = delta; map[i][j].beta = beta;
	prev_map_activity[i][j]=map_activity[i][j]=0.0;
      }

  srand48(seed);    /*        start random number sequence                  */
  startt=0;         /* startt may be changed by iterate_snapshots           */

/* Allocate the weights array in each neuron. Do only for the number of
 * neurons we are using to save space.
 */
  Mem_Once_Alloced = True;  /* To know that memory has been allocated
                             * once and may need to be freed later.         */
  for (i=0; i<N; i++)
    for(j=0; j<N; j++){
      allocsize = sizeof(double)*(lat_exc_dimension)     ;   
      map[i][j].lat_exc_wts = (double*)alloce(allocsize) ;

      allocsize = sizeof(double)*(lat_inh_dimension)     ;   
      map[i][j].lat_inh_wts = (double*)alloce(allocsize) ;

      allocsize = sizeof(double)*(input_dimension+1)     ; 
      map[i][j].weights = (double*)alloce(allocsize)     ;
    }
  input_vectors = (double*)alloce(allocsize)  ;
  raw_input     = (double*)alloce(allocsize)  ;

#ifdef XWINDOW
  display_parameters(&disp);
#endif
  return(1);
}
/* ---------------------------------------------------------------------- */

void read_params(fp) /* Reads the simulation parameter file               */
FILE *fp;
{
  char s[100];
  int i,k,l,killphase;

  fgets(s,99,fp); fgets(s,99,fp); /* First 2 lines for title of the file. */


  fscanf(fp,"%d", &input_dimension); fgets(s,99,fp);

  /* -------Read the range and randomness for each 'real' input dimension.*/

  input_ranges  = (Ranges*)alloce(sizeof(Ranges)*input_dimension);
  for(i=0; i<input_dimension; i++) 
    fscanf(fp,"%lf %lf %lf", &(input_ranges[i].low), &(input_ranges[i].high),
	   &(input_ranges[i].randomness));
  fgets(s,99,fp);

  fscanf(fp,"%d",&distribution); fgets(s,99,fp);
  fscanf(fp,"%d %d %d %d", &N, &border,&dynamics, &symmetric_exc_wts); 
  fgets(s,99,fp);
  fscanf(fp,"%d %d %d", &displaying,&disp_rate,&seed); fgets(s,99,fp);
  fscanf(fp,"%d", &tend)                             ; fgets(s,99,fp);
  fscanf(fp,"%lf %lf", &lat_exc_randomness, &lat_inh_randomness); fgets(s,99,fp);
  fscanf(fp,"%lf %lf %lf",&alpha_delta, &alpha_beta, &act_dep_frac);
  fgets(s,99,fp);
  fscanf(fp,"%lf %lf", &final_delta, &final_beta);
  fgets(s,99,fp);

  fscanf(fp,"%d %lf %lf", &nkillphases, &exc_death, &inh_death);fgets(s,99,fp);
  for (killphase=0; killphase < nkillphases; killphase++)
    fscanf(fp,"%d",  &kill_phases[killphase]);
  fgets(s,99,fp);

  fscanf (fp,"%d %d %d %lf %lf %lf", &(tsettle)    , 
	                          &(exc_rad)    , 
	                          &(inh_rad)    ,
	                          &(alpha_input),
	                          &(alpha_exc)  ,
	                          &(alpha_inh) ); fgets(s,99,fp);
  fscanf (fp,"%lf %lf %lf %lf",  &(delta), &(beta), 
	                         &(gammaexc)  , 
	                         &(gammainh) ); fgets(s,99,fp);

  fscanf(fp,"%d", &snapshots[0])    ;
  for(i=0; i<MAXSNAPS && snapshots[i]<SNAPSHOTEND; i++)
    fscanf(fp,"%d", &(snapshots[i+1]));
  fgets(s,99,fp)                    ;
  nextsnapshot=0                    ;
}
/* ---------------------------------------------------------------------- */
void init_weights()
{
/* init_weights and distr_weight depend on the way you want to distribute
 * the initial weights. However the function tocartesian() is a general
 * function that accepts a set of initial weights in a multidimensional
 * 'angle space' and converts them to normalized cartesian coordinates.
 */
  double winit[MAXDIM-1]    ;
  int i,j,k; 
  double center,low,high;

  /* Initialize input weights within the given degree of randomness
   * retinotopically.
   */
  for(i=border; i<N-border; i++)
    for(j=border; j<N-border; j++)
      {
	winit[0]=distr_weight(0,i-border,(N-(2*border)),
			      input_ranges[0].randomness);
	winit[1]=distr_weight(1,j-border,(N-(2*border)),
			      input_ranges[1].randomness);

	/* Initialize the rest of the components randomly within their
         * specified ranges.
	 */
	for(k=2; k < input_dimension; k++)
	  /* Distribute about center of range with given randomness */
	  
	  winit[k] = scalelu( k, input_ranges[k].randomness * drand48() );
	
	tocartesian(winit,map[i][j].weights);/*To normalized cartesian coords*/
      }

  if (symmetric_exc_wts)
    setup_latw_symmetric();/* Output lateral exc wts at border are symmetric */
  else setup_latw_unsymmetric();
}

/* ---------------------------------------------------------------------- */
double distr_weight(component,position, maxneurons, radius)
int component, position, maxneurons;
double radius;
{
  double inp, low, high; double temp;

  inp =((double)position+1.0)/((double)maxneurons+1.0);
                                     /* The weights can lie anywhere  */
  low=inp-radius; high=inp+radius;   /* between low and high.              */
  if (low<0.0) low=0.0           ;
  if (high>1.0) high=1.0         ;

  temp = scalelu( component, low+( (high-low)*drand48() )
		  );
  return(temp);
}
/* ---------------------------------------------------------------------- */
void tocartesian(winit,normalized_vector)
double winit[MAXDIM-1]   ;
double *normalized_vector; /* Of length input_dimension+1 */
{
/* This function accepts a 'input_dimension' angles and converts them to
 * normalized cartesian values in 'input_dimension+1' space. It initializes
 * the input weights of neuron (i,j) with these values.
 */
  int k; double cosproduct;

  for(k=0,cosproduct=1; k<input_dimension; k++)
    {
      normalized_vector[k] = sin(winit[k]) * cosproduct;
      cosproduct *= cos(winit[k])                      ;
    }
  normalized_vector[input_dimension] = cosproduct      ;
#ifdef DEBUG
  printf("=>Cosproduct: no.of coses=%d\n",k);
#endif

}
/* ---------------------------------------------------------------------- */
void setup_latw_unsymmetric()  
{
/* Sets up lateral weights of neurons based on where they are in the map.
 * Neurons at the border have fewer connections than at the center.
 * Note that inhibition wts are initialized to positive.
 * The lateral wts are distributed around the mean with the given randomness.
 * The lowk,lowl,highk,highl variables keep track of the spread of connections.
 * Connections can at most spread to both borders of the network.
 */
 int i,j,k,l,m,exc_ar_width, inh_ar_width;
 int no_of_inh_connections,no_of_exc_connections;
 int elowk,ehighk,ilowk,ihighk,elowl,ehighl,ilowl,ihighl;
 double meanexc_wt, meaninh_wt; /* Mean weight of neuron's connection */

#ifdef DEBUG1
 double totale,totali; 
#endif

 exc_ar_width = 2*exc_rad+1;
 inh_ar_width = 2*inh_rad+1;

 fprintf(runtimefile,"*=========== \t Lateral weight statistics \t============*\n");
 fprintf(runtimefile,"*== \t (i,j) \t\t mean_exc_wt*gammaexc \t mean_inh_wt*gammainh \n");
 for(i=0; i<N; i++){
   elowk  = i-exc_rad; if (elowk < 0) elowk =0    ;
   ehighk = i+exc_rad; if (ehighk>=N) ehighk = N-1;
   ilowk  = i-inh_rad; if (ilowk < 0) ilowk =0    ;
   ihighk = i+inh_rad; if (ihighk>=N) ihighk = N-1;
   for(j=0; j<N; j++)
     { 
       elowl  = j-exc_rad; if (elowl < 0) elowl =0    ;
       ehighl = j+exc_rad; if (ehighl>=N) ehighl = N-1;
       ilowl  = j-inh_rad; if (ilowl < 0) ilowl =0    ;
       ihighl = j+inh_rad; if (ihighl>=N) ihighl = N-1;

       no_of_exc_connections = (ehighk-elowk+1)*(ehighl-elowl+1);
       no_of_inh_connections = (ihighk-ilowk+1)*(ihighl-ilowl+1);
       meanexc_wt = 1.0/(double)(no_of_exc_connections);
       meaninh_wt = 1.0/(double)(no_of_inh_connections);

/* Print some statistics to the runtime file on weight distributions */
       fprintf(runtimefile,"*--\t (%d,%d) \t\t %.6lf \t %.6lf \n",
	       i,j,meanexc_wt*gammaexc,meaninh_wt*gammainh);

       for(m=0; m<lat_exc_dimension; m++)  /* Initialize to zero */
	 map[i][j].lat_exc_wts[m]=0.0;
	    
       for(k=elowk; k<=ehighk; k++)        /* Init excitation weights */
	 for(l=elowl; l<=ehighl; l++)
	   map[i][j].lat_exc_wts[(k-i+exc_rad)*exc_ar_width + 
				 (l-j+exc_rad) ] = 
				   dist_lat_wt(meanexc_wt,
					       lat_exc_randomness);

       for(m=0; m<lat_inh_dimension; m++)  /* Initialize to zero */
	 map[i][j].lat_inh_wts[m]=0.0;

       for(k=ilowk; k<=ihighk; k++)        /* Init inhibition weights */
	 for(l=ilowl; l<=ihighl; l++)
	   map[i][j].lat_inh_wts[(k-i+inh_rad)*inh_ar_width + 
				 (l-j+inh_rad) ] = 
				   dist_lat_wt(meaninh_wt,
					       lat_inh_randomness);
#ifdef DEBUG1
       printf("** Lat wt init: (i,j) = (%d,%d)\n",i,j);
       printf("** elowk=%d, ehighk=%d, ilowk=%d, ihighk=%d, elowl=%d, ehighl=%d, ilowl=%d, ihighl=%d\n",elowk,ehighk,ilowk,ihighk,elowl,ehighl,ilowl,ihighl);
       printf("** no_of_exc_connections=%d,no_of_inh_connections=%d\n",
	      no_of_exc_connections,no_of_inh_connections);
       totale=totali=0.0;
       for(m=0; m<lat_exc_dimension; m++)  
	 totale+=map[i][j].lat_exc_wts[m];
       for(m=0; m<lat_inh_dimension; m++)  
	 totali+=map[i][j].lat_inh_wts[m];
       printf("** total exc_wt= %lf, total inh_wt= %lf\n",totale,totali);
#endif

     }
 }
 fprintf(runtimefile,"*=================================================*\n");
}
/* ---------------------------------------------------------------------- */
void setup_latw_symmetric()  
{
/* Sets up lateral weights of neurons based on where they are in the map.
 * Neurons at the border have fewer connections than at the center.
 * The excitatory weights are symmetric at the border.
 */
 int i,j,k,l,m,exc_ar_width, inh_ar_width;
 int no_of_inh_connections,no_of_exc_connections;
 int elowk,ehighk,ilowk,ihighk,elowl,ehighl,ilowl,ihighl;
 double meanexc_wt, meaninh_wt; /* Mean weight of neuron's connection */

#ifdef DEBUG1
 double totale,totali; 
#endif

 exc_ar_width = 2*exc_rad+1;
 inh_ar_width = 2*inh_rad+1;

 fprintf(runtimefile,"*=========== \t Lateral weight statistics \t============*\n");
 fprintf(runtimefile,"*== \t (i,j) \t\t mean_exc_wt*gammaexc \t mean_inh_wt*gammainh \n");
 for(i=0; i<N; i++){
   elowk  = i-exc_rad; if (elowk < 2*exc_rad) elowk = (i+1) >> 1;
   ehighk = i+exc_rad; if (ehighk>=N- 2*exc_rad) ehighk = N-1 -((N-i) >> 1);
   ilowk  = i-inh_rad; if (ilowk < 0) ilowk =0    ;
   ihighk = i+inh_rad; if (ihighk>=N) ihighk = N-1;
   for(j=0; j<N; j++)
     { 
       elowl  = j-exc_rad; if (elowl < 2*exc_rad) elowl = (j+1) >>1 ;
       ehighl = j+exc_rad; if (ehighl>=N-2*exc_rad) ehighl = N-1 - ((N-j) >>1);
       ilowl  = j-inh_rad; if (ilowl < 0) ilowl =0    ;
       ihighl = j+inh_rad; if (ihighl>=N) ihighl = N-1;

       no_of_exc_connections = (ehighk-elowk+1)*(ehighl-elowl+1);
       no_of_inh_connections = (ihighk-ilowk+1)*(ihighl-ilowl+1);
       meanexc_wt = 1.0/(double)(no_of_exc_connections);
       meaninh_wt = 1.0/(double)(no_of_inh_connections);

/* Print some statistics to the runtime file on weight distributions */
       fprintf(runtimefile,"*--\t (%d,%d) \t\t %.6lf \t %.6lf \n",
	       i,j,meanexc_wt*gammaexc,meaninh_wt*gammainh);

       for(m=0; m<lat_exc_dimension; m++)  /* Initialize to zero */
	 map[i][j].lat_exc_wts[m]=0.0;
	    
       for(k=elowk; k<=ehighk; k++)        /* Init excitation weights */
	 for(l=elowl; l<=ehighl; l++)
	   map[i][j].lat_exc_wts[(k-i+exc_rad)*exc_ar_width + 
				 (l-j+exc_rad) ] = 
				   dist_lat_wt(meanexc_wt,
					       lat_exc_randomness);

       for(m=0; m<lat_inh_dimension; m++)  /* Initialize to zero */
	 map[i][j].lat_inh_wts[m]=0.0;

       for(k=ilowk; k<=ihighk; k++)        /* Init inhibition weights */
	 for(l=ilowl; l<=ihighl; l++)
	   map[i][j].lat_inh_wts[(k-i+inh_rad)*inh_ar_width + 
				 (l-j+inh_rad) ] = 
				   dist_lat_wt(meaninh_wt,
					       lat_inh_randomness);
#ifdef DEBUG1
       printf("** Lat wt init: (i,j) = (%d,%d)\n",i,j);
       printf("** elowk=%d, ehighk=%d, ilowk=%d, ihighk=%d, elowl=%d, ehighl=%d, ilowl=%d, ihighl=%d\n",elowk,ehighk,ilowk,ihighk,elowl,ehighl,ilowl,ihighl);
       printf("** no_of_exc_connections=%d,no_of_inh_connections=%d\n",
	      no_of_exc_connections,no_of_inh_connections);
       totale=totali=0.0;
       for(m=0; m<lat_exc_dimension; m++)  
	 totale+=map[i][j].lat_exc_wts[m];
       for(m=0; m<lat_inh_dimension; m++)  
	 totali+=map[i][j].lat_inh_wts[m];
       printf("** total exc_wt= %lf, total inh_wt= %lf\n",totale,totali);
#endif

     }
 }
 fprintf(runtimefile,"*=================================================*\n");
}
/* ---------------------------------------------------------------------- */ 
double dist_lat_wt(value, randomness)
double value, randomness; /* Randomness in range 0-1 */
{
/* This function returns a value randomly distributed within a range
 * value*randomness on either side of value 
 */
  double low, add;

  low = value * (1.0 - randomness)  ;
  add = 2*value*randomness*drand48();

  return(low+add);
}
/* ---------------------------------------------------------------------- */ 
/*                         I/O and miscellaneous functions                */ 
/* ---------------------------------------------------------------------- */

int save_current_binary()
{
/* Appends snapshot to simufile.bin. The output is a binary image of the 
 * list of weights. The program may use either this procedure or the previous
 * one to save weights.
 */
  FILE *fp              ;
  char buf[50]          ; 
  extern size_t fwrite();

  sprintf(buf, "%s.bin",simufile)                     ;
  if ((fp=fopen(buf,"ab"))==NULL){
    fprintf(stderr,"The simulation file %s couldn't be opened for appending\n",
	    buf)                                      ;
    return(0)                                         ;
  }   
  if (fprintf(fp,"%d", t)==EOF)
    fprintf(stderr,"Error appending to file\n")       ;
  weights_IO(fwrite,fp)                               ;
  fclose(fp)                                          ;

  nextsnapshot++                                      ;
  return(1)                                           ;
}  
/* ---------------------------------------------------------------------- */
/* This procedure does reading and writing of weights to a binary file. fp
 * is a binary file pointer.
 */
void weights_IO(dofun,fp)
size_t (*dofun)();
FILE *fp;
{
  int i,j;
  size_t nobj;

  for(i=0; i<N; i++)
    for(j=0; j<N; j++){
      if((nobj=(*dofun)((void*)(& map[i][j].delta), sizeof(double), 
			1, fp)) != 1)
	fprintf(stderr,"Error in IO - delta. nobj = %d\n",nobj);

      if((nobj=(*dofun)((void*)(& map[i][j].beta), sizeof(double), 
			1, fp)) != 1)
	fprintf(stderr,"Error in IO - beta. nobj = %d\n",nobj);

      if((nobj=(*dofun)((void*)map[i][j].weights, sizeof(double), 
			(input_dimension+1), fp)) != (input_dimension+1))
	fprintf(stderr,"Error in IO - input wts. nobj = %d\n",nobj);

      if((nobj=(*dofun)((void*)map[i][j].lat_exc_wts, sizeof(double), 
			lat_exc_dimension, fp)) != lat_exc_dimension)
	fprintf(stderr,"Error in IO-lat exc wts.\n");

      if((nobj=(*dofun)((void*)map[i][j].lat_inh_wts, sizeof(double), 
			lat_inh_dimension, fp)) != lat_inh_dimension)
	fprintf(stderr,"Error in IO - lat inh wts.\n");
    }
}

/* ---------------------------------------------------------------------- */
void iterate_weights(dofun) 
int (*dofun)();
/* Uses dofun on the weights of each neuron. This function can be convenient
 * for writes and reads if you are saving weights in ascii format. Right now,
 * it is only used in iterate_snapshots to update the random number generator.
 * writefun(),readfun() and randfun() go with this function.
 */
{
   int i,j,k;
  
  for(i=0; i<N; i++)
    for(j=0; j<N; j++){

      for(k=0; k<=input_dimension; k++)
	(*dofun)(&map[i][j].weights[k]);
      for(k=0; k< lat_exc_dimension; k++)
	(*dofun)(&map[i][j].lat_exc_wts[k]);
      for(k=0; k< lat_inh_dimension; k++)
	(*dofun)(&map[i][j].lat_inh_wts[k]);
    }  
 }
/* ---------------------------------------------------------------------- */
void readfun(place)
double *place;
{
  if (fscanf(fp,"%lf", place)==EOF)
    fprintf(stderr,"EOF reached on read from file\n");
} 
/* ---------------------------------------------------------------------- */
void writefun(place)
double *place;
{
  if (fprintf(fp,"%lf\n", *place)==EOF) 
    fprintf(stderr,"Error writing to file\n");
;
} 
/* ---------------------------------------------------------------------- */
void randfun(place)
double *place;
{
  *place = drand48();  /* Rand value in [0.0, 1.0).  */
} 

/* ---------------------------------------------------------------------- *
 *                        Housekeeping Routines                           *
 * ---------------------------------------------------------------------- */

#ifdef XWINDOW
void add2history(raw_input) /* The latest raw_input is at head of list */
double *raw_input;
{
  Input_history *new; int i;

  new = (Input_history*)alloce(sizeof(Input_history));
  new->input_vector = (double*)alloce(sizeof(double)*(input_dimension+1));
  new->next = input_list; /* Attach the head of the input list */
  input_list = new      ;
  
  for(i=0; i<=input_dimension; i++)
    input_list->input_vector[i] = raw_input[i];
}
/* ------------------------------------------------------------------------ */
void free_inptlist() /* Frees all the input data stored */
{
 Input_history *prev,*current;

 for(prev=current=input_list; current!=NULL; )
   {
     free( (char*)(prev->input_vector));
     current = prev->next;
     free((char*)prev);
     prev=current;
   }
 input_list = NULL;
}
#endif

/* ------------------------------------------------------------------------ */
void free_old_memory() /* Gets called only if Mem_Once_Alloced is True */
{
  int i, j;

#ifdef XWINDOW
  free_inptlist();       /* Points are saved only for display */
#endif

  for (i=0; i<N; i++)
    for(j=0; j<N; j++){
       free((char *) map[i][j].weights);
       map[i][j].weights =NULL         ;
       free((char *) map[i][j].lat_exc_wts);
       free((char *) map[i][j].lat_inh_wts);
       map[i][j].lat_exc_wts=map[i][j].lat_inh_wts=NULL;
     }
  free((char *)input_vectors           );
  free((char *)raw_input               );
  free((char *)input_ranges            );
}
/* ------------------------------------------------------------------------ */
/*                          Parameter IO routines                           */
/* ------------------------------------------------------------------------ */
void store_in_bboard(blackboard,index,name,type,flag,pointer)
Blackboard *blackboard;char *name; int type,flag; char *pointer;
{
  strcpy(blackboard->list[index].param_name, name);
  blackboard->list[index].data_type  = type       ;
  blackboard->list[index].update_flag    = flag   ;
  
  switch(type){
  case CHAR  : blackboard->list[index].data_ptr.c = pointer        ; break;
  case INT   : blackboard->list[index].data_ptr.i = (int*)pointer  ; break;
  case FLOAT : blackboard->list[index].data_ptr.f = (double*)pointer; break;
  }
  blackboard->no_of_entries++;
}
/* ------------------------------------------------------------------------ */
void setup_blackboard(blackboard)
Blackboard *blackboard;
/* Routine that sets up the interface of lcsofm with other output routines.
 * The blackboard is used to provide a standard interface to interactive
 * displays, which may vary from one implementation to another. The interface
 * can directly change the variables that are changeable by aliasing through
 * the pointer provided.
 */
{
  blackboard->no_of_entries=0;
  store_in_bboard(blackboard,0, "iterations",  INT  ,False,(char*)&t)         ;
  store_in_bboard(blackboard,1, "tend",        INT  ,True,(char*)&tend)       ;
  store_in_bboard(blackboard,2, "disp_rate",   INT  ,True,(char*)&disp_rate)  ;
  store_in_bboard(blackboard,3, "step_through",INT ,True,(char*)&step_through);
  store_in_bboard(blackboard,4, "exc_rad",     INT  ,False,(char*)&exc_rad)   ;
  store_in_bboard(blackboard,5, "inh_rad",     INT  ,False,(char*)&inh_rad)   ;
  store_in_bboard(blackboard,6, "tsettle",     INT  ,True,(char*)&tsettle)    ;
  store_in_bboard(blackboard,7, "gammaexc",    FLOAT,True,(char*)&gammaexc)   ;
  store_in_bboard(blackboard,8, "gammainh",    FLOAT,True,(char*)&gammainh)   ;
  store_in_bboard(blackboard,9, "alpha_input", FLOAT,True,(char*)&alpha_input);
  store_in_bboard(blackboard,10,"alpha_exc",   FLOAT,True,(char*)&alpha_exc)  ;
  store_in_bboard(blackboard,11,"alpha_inh",   FLOAT,True,(char*)&alpha_inh)  ;
  store_in_bboard(blackboard,12,"alpha_delta", FLOAT,True,(char*)&alpha_delta);
  store_in_bboard(blackboard,13,"alpha_beta",  FLOAT,True,(char*)&alpha_beta) ;
  store_in_bboard(blackboard,14,"act_dep_frac",FLOAT,True,(char*)&act_dep_frac);
  store_in_bboard(blackboard,15,"initial delta",FLOAT,True,(char*)&delta)     ;
  store_in_bboard(blackboard,16,"initial beta", FLOAT,True,(char*)&beta)      ;
  store_in_bboard(blackboard,17,"final_delta", FLOAT,True,(char*)&final_delta);
  store_in_bboard(blackboard,18,"final_beta",  FLOAT,True,(char*)&final_beta) ;
  store_in_bboard(blackboard,19,"exc_death",   FLOAT,True,(char*)&exc_death)  ;
  store_in_bboard(blackboard,20,"inh_death",   FLOAT,True,(char*)&inh_death)  ;
  store_in_bboard(blackboard,21,"kill_phases", INT  ,True,(char*)kill_phases) ;
  store_in_bboard(blackboard,22,"nkillphases", INT ,False,(char*)&nkillphases);
  store_in_bboard(blackboard,23, "snapshots",  INT  ,True,(char*)snapshots)   ;
  store_in_bboard(blackboard,24,"dynamics",    INT  ,True,(char*)&dynamics)   ;

}
/* ------------------------------------------------------------------------ */
void print_params(fp,blackboard)
FILE *fp; Blackboard *blackboard;
{
/* Prints parameter values to a file from the parameter blackboard.
 */
  int i;

  fprintf(fp,"**** Current Parameters ****\n");
  for(i=0; i< blackboard->no_of_entries; i++){

    fprintf(fp,"%s:",blackboard->list[i].param_name);
    switch(blackboard->list[i].data_type){
    case CHAR :fprintf(fp,"%c\n",*(blackboard->list[i].data_ptr.c)); break;
    case INT  :fprintf(fp,"%d\n",*(blackboard->list[i].data_ptr.i)); break;
    case FLOAT:fprintf(fp,"%lf\n",*(blackboard->list[i].data_ptr.f)); break;
    }
  }
  fprintf(fp,"\n");
}
/* ------------------------------------------------------------------------ */
