/* ---------------------------------------------------------------------- 
 *                  visualize.c: May 4, 1993.
 * Corresponds to the program xlcsofm
 *
 * The program can produce output in two formats. one for mathematica
 * and the other for maplev. four sets of data are output - the 
 * excitation weights, inhibition weights, the mexican hat, and the
 * distribution of mex hat peaks over the grid surface.
 *
 *
 * future work:
 *       1. produce avs data sets
 *       2. produce as another output a grid with mex hats jutting out
 *          at intersting points.
 * ---------------------------------------------------------------------- */

#include <stdlib.h>
#include "globals.h"

#define MATH           1
#define MAPLE          2

#define HORIZ          1
#define VERT           2

int  init_params()                                ;
void modify_lat_wts()                             ;
void read_params()                                ;
void init_files()                                 ;
void menu()                                       ;
void dumpij()                                     ;
void compute_mexhat()                             ;
void output()                                     ;
void maple()                                      ;
void mathematica()                                ;
void dumppeaks()                                  ;
void weights_IO()                                 ;
void free_all()                                   ;
void dump_zwts()                                  ;
int alloc_exc_size, alloc_inh_size, alloc_inp_size;
long snap_to_snapoffset, neuron_offset            ;
int output_routine, snap_index, scanline          ;
int step, no_of_neurons,starti,startj             ;
int snapshot;
char orientation[20] ;   /* contains "horizontal" or "vertical" */

FILE *wtfile, *excfile, *inhfile, *mexfile, *peakfile, *deltafile,
  *betafile, *zfile;

struct lookup{
  int snapshot;
  long offset;
}lookup_snaps[MAXSNAPS];

Neuron Temp;  /* to hold the exc and inh wts. */

/* ---------------------------------------------------------------------- */

void main(argc, argv)
int argc;
char **argv;
{

  if (!init_params(argv[1])) exit(0)   ;    /* file not found */
  menu();                                   /* get user choices */
  init_files();                             /* open and write header files */
  output(starti,startj);                    /* dump the weights */
  free_all();
} 


/* ---------------------------------------------------------------------- */

int init_params(argv)
char *argv;
{
  int c,i,j,allocsize,temp; 
  FILE *fp;
  extern char* alloce();
  char wtfil[100];


  sprintf(simufile, "%s", argv);
  if ( (fp=fopen(simufile,"r"))==NULL){
    fprintf(stderr, "file %s not found\n", simufile);
    return(0);
  }
  read_params(fp)  ;
  fclose(fp)       ;  /* close off the parameter file */

  lat_exc_dimension = (2*exc_rad + 1) * (2*exc_rad + 1);
  lat_inh_dimension = (2*inh_rad + 1) * (2*inh_rad + 1);

  sprintf(wtfil,"%s.bin",simufile);
  if( (wtfile=fopen(wtfil,"rb"))==NULL)
    {
      fprintf(stderr,"wt file %s could not be opened\n",wtfil);
      return(0);
    }

  for (i=0; i<N; i++)
    for(j=0; j<N; j++){

      allocsize = sizeof(double)*(input_dimension+1)     ; 
      map[i][j].weights = (double*)alloce(allocsize)    ;

      allocsize = sizeof(double)*(lat_exc_dimension)     ;   
      map[i][j].lat_exc_wts = (double*)alloce(allocsize);

      allocsize = sizeof(double)*(lat_inh_dimension)     ;   
      map[i][j].lat_inh_wts = (double*)alloce(allocsize);

    }
  /* Use prev allocsize to allocate these inh wts for mex hat storage */
  Temp.lat_inh_wts = (double*)alloce(allocsize); 

  return(1);
}
/* ---------------------------------------------------------------------- */
char* alloce(allocsize)/* allocator with error check for mem unavailability */
int allocsize;
{
  char *new;

  if (( new = malloc(allocsize))==NULL){
    fprintf(stderr,"not enough memory. exiting\n");
    exit(0);
  }
  return(new);
}
/* ---------------------------------------------------------------------- */

void read_params(fp) /* Reads the simulation parameter file               */
FILE *fp;
{
  char s[100];
  int i,k,l,killphase;

  fgets(s,99,fp); fgets(s,99,fp); /* First 2 lines for title of the file. */


  fscanf(fp,"%d", &input_dimension); fgets(s,99,fp);

  /* -------Read the range and randomness for each 'real' input dimension.*/

  input_ranges  = (Ranges*)alloce(sizeof(Ranges)*input_dimension);
  for(i=0; i<input_dimension; i++) 
    fscanf(fp,"%lf %lf %lf", &(input_ranges[i].low), &(input_ranges[i].high),
	   &(input_ranges[i].randomness));
  fgets(s,99,fp);

  fscanf(fp,"%d",&distribution); fgets(s,99,fp);
  fscanf(fp,"%d %d %d", &N, &border,&dynamics, &symmetric_exc_wts) ; fgets(s,99,fp);
  fscanf(fp,"%d %d %d", &displaying,&disp_rate,&seed); fgets(s,99,fp);
  fscanf(fp,"%d", &tend)                           ; fgets(s,99,fp);
  fscanf(fp,"%lf %lf", &lat_exc_randomness, &lat_inh_randomness); fgets(s,99,fp);
  fscanf(fp,"%lf %lf %lf",&alpha_delta, &alpha_beta, &act_dep_frac);
  fgets(s,99,fp);
  fscanf(fp,"%lf %lf", &final_delta, &final_beta);
  fgets(s,99,fp);

  fscanf(fp,"%d %lf %lf", &nkillphases, &exc_death, &inh_death);fgets(s,99,fp);
  for (killphase=0; killphase < nkillphases; killphase++)
    fscanf(fp,"%d",  &kill_phases[killphase]);
  fgets(s,99,fp);

  fscanf (fp,"%d %d %d %lf %lf %lf", &(tsettle)    , 
	                          &(exc_rad)    , 
	                          &(inh_rad)    ,
	                          &(alpha_input),
	                          &(alpha_exc)  ,
	                          &(alpha_inh) ); fgets(s,99,fp);
  fscanf (fp,"%lf %lf %lf %lf",  &(delta), &(beta), 
	                         &(gammaexc)  , 
	                         &(gammainh) ); fgets(s,99,fp);

  fscanf(fp,"%d", &snapshots[0])    ;
  for(i=0; i<MAXSNAPS && snapshots[i]<SNAPSHOTEND; i++)
    fscanf(fp,"%d", &(snapshots[i+1]));
  fgets(s,99,fp)                    ;
  nextsnapshot=0                    ;
}
/* ---------------------------------------------------------------------- */
void init_files()
/* open the output files and put header text
 */
{
  char bufexc[100],bufinh[100],bufmex[100],
  bufpeaks[100],bufdelta[100],bufbeta[100],bufz[100];

  switch(output_routine){             /* set up correct extensions */

  case MATH: /* mathematica */
    sprintf(bufexc,"%s.%d.ex.math",simufile,snapshot);
    sprintf(bufinh,"%s.%d.in.math",simufile,snapshot);
    sprintf(bufmex,"%s.%d.mx.math",simufile,snapshot);
    sprintf(bufpeaks,"%s.%d.p.math",simufile,snapshot);
    sprintf(bufdelta,"%s.%d.d.math",simufile,snapshot);
    sprintf(bufbeta,"%s.%d.b.math",simufile,snapshot);
    sprintf(bufz,"%s.%d.z.math",simufile,snapshot);
    break;

  case MAPLE: /* maple */
    sprintf(bufexc,"%s.%d.ex.mpl",simufile,snapshot);
    sprintf(bufinh,"%s.%d.in.mpl",simufile,snapshot);
    sprintf(bufmex,"%s.%d.mx.mpl",simufile,snapshot);
    sprintf(bufpeaks,"%s.%d.p.mpl",simufile,snapshot);
    sprintf(bufdelta,"%s.%d.d.mpl",simufile,snapshot);
    sprintf(bufbeta,"%s.%d.b.mpl",simufile,snapshot);
    sprintf(bufz,"%s.%d.z.mpl",simufile,snapshot);
    break;
  }
  if( (excfile=fopen(bufexc,"w"))==NULL ){
      fprintf(stderr,"out file %s could not be opened\n",bufexc);
      exit(0);
    }
  if( (inhfile=fopen(bufinh,"w"))==NULL ){
    fprintf(stderr,"out file %s could not be opened\n",bufinh);
    exit(0);
  } 
  if( (mexfile=fopen(bufmex,"w"))==NULL ){
    fprintf(stderr,"out file %s could not be opened\n",bufmex);
    exit(0);

  }
  if( (peakfile=fopen(bufpeaks,"w"))==NULL){
    fprintf(stderr,"out file %s could not be opened\n",bufpeaks);
    exit(0);
  }
  if( (deltafile=fopen(bufdelta,"w"))==NULL){
    fprintf(stderr,"out file %s could not be opened\n",bufdelta);
    exit(0);
  }
  if( (betafile=fopen(bufbeta,"w"))==NULL){
    fprintf(stderr,"out file %s could not be opened\n",bufbeta);
    exit(0);
  }
  if( (zfile=fopen(bufz,"w"))==NULL){
    fprintf(stderr,"out file %s could not be opened\n",bufz);
    exit(0);
  }
}
/* ---------------------------------------------------------------------- */
void menu()
{
  int i;

  printf("\nenter the output routine(mathematica - 1, maple - 2. default 1)=>");
  scanf("%d",&output_routine);
  if (output_routine !=1) output_routine=2; /* default */
  printf("\n--routine: %d\n", output_routine);

  printf("enter snapshot=>");
  scanf("%d", &snapshot);
  for(i=0; i<MAXSNAPS && lookup_snaps[i].snapshot < SNAPSHOTEND &&
      lookup_snaps[i].snapshot < snapshot; i++);
  printf("\n--snap: %d\n", snapshot);

  printf("enter starting neuron (i,j). default (0,0)=>");
  scanf("%d %d", &starti, &startj);
  if ( starti >= N || startj >= N){
    starti=0; startj=0;
  }
  printf("\n--(starti, startj): (%d,%d)\n",starti, startj);

  printf("enter choice of scanline.( horiz - 1, vert - 2. default 1)=>");
  scanf("%d",&scanline); 
  if (scanline == 2)                   /* set the message buffer */
    sprintf(orientation,"vertical");
  else{
    sprintf(orientation,"horizontal");
    scanline =1;      /* default */
  } 
  printf("\n--scanline: %s\n",orientation);

  printf("enter choice of step along scanline. default 1=>");
  scanf("%d", &step); 
  if ((step < 0) || (step >= N)) step = 1;
  printf("\n--step: %d\n",step);

  printf("enter no. of neurons along scanline. default 1=>");
  scanf("%d",&no_of_neurons); 
  if ( (no_of_neurons < 1) || (no_of_neurons >= N) ) no_of_neurons = 1;
  printf("\n--no of neurons: %d\n",no_of_neurons);

}
/* ---------------------------------------------------------------------- */
void dumpsigmoid(deltafile,betafile)
FILE *deltafile, *betafile;
{
  int i,j,k;
  int exc_ar_width, inh_ar_width;
  double peak;

  exc_ar_width = 2*exc_rad+1;
  inh_ar_width = 2*inh_rad+1;

  switch(output_routine){
  case MATH:
    for(i=0; i<N; i++)
      {
	for(j=0; j<N-1; j++){
	  fprintf(deltafile,"%lf", map[i][j].delta);
	  fprintf(betafile,"%lf", map[i][j].beta);
	}
	fprintf(deltafile,"%lf\n",map[i][N-1].delta);
	fprintf(betafile,"%lf\n", map[i][N-1].beta);
      }
    break;

  case MAPLE:
  /* Print header for maple */
    fprintf(deltafile,"with(linalg);\nwith(plots);\n");
    fprintf(deltafile,"z := matrix( [ \n");
    fprintf(betafile,"with(linalg);\nwith(plots);\n");
    fprintf(betafile,"z := matrix( [ \n");

    for(i=0; i<N; i++){
      fprintf(deltafile, "[");
      fprintf(betafile, "[");

      for(j=0; j<N-1; j++){
	fprintf(deltafile,"%lf,", map[i][j].delta );
	fprintf(betafile,"%lf,", map[i][j].beta );
      }
      fprintf(deltafile,"%lf]",map[i][N-1].delta);
      fprintf(betafile,"%lf]",map[i][N-1].beta);

      if (i==N-1){
	fprintf(deltafile," ] );\n");
	fprintf(betafile," ] );\n");
      }
      else{
	fprintf(deltafile,",\n");
	fprintf(betafile,",\n");
      }
    }
    fprintf(deltafile,"matrixplot(z,title=`LOWER THRESHOLD(DELTA)`,labels=[`x`,`y`,`d`],axes=BOXED,style=PATCH,shading=Z);\n");
    fprintf(betafile,"matrixplot(z,title=`UPPER THRESHOLD(BETA)`,labels=[`x`,`y`,`d`],axes=BOXED,style=PATCH,shading=Z);\n");
    break;
  }
}

/* ---------------------------------------------------------------------- */
void dumpij(i,j,snapoffset,neuron_offset) 
int i,j;
long snapoffset,neuron_offset;
{
  int m;
  size_t nobj;
  size_t read_offset; char title[100],zaxis[100];


  printf("gammaexc: %.3lf,gammainh: %.3lf\n",gammaexc,gammainh);

  for(m=0; m<lat_exc_dimension; m++)   
    map[i][j].lat_exc_wts[m] *= gammaexc;
  for(m=0; m<lat_inh_dimension; m++)
    map[i][j].lat_inh_wts[m] *= -gammainh;

  switch(output_routine){

  case MATH: 
    mathematica(excfile,map[i][j].lat_exc_wts,lat_exc_dimension,exc_rad,i,j);
    mathematica(inhfile,map[i][j].lat_inh_wts,lat_inh_dimension,inh_rad,i,j);
    break;

  case MAPLE: 
    sprintf(zaxis,"Exc");
	   
    sprintf(title,"EXCITATION WTS OF A NEURON AT (%d,%d) IN A %dx%d GRID",
	    i,j,N,N);
    maple(excfile,map[i][j].lat_exc_wts,lat_exc_dimension,exc_rad,i,j,
	  title,zaxis );
    sprintf(title,"INHIBITION WTS OF A NEURON AT (%d,%d) IN A %dx%d GRID",
	    i,j,N,N);
    sprintf(zaxis,"Inh");
    maple(inhfile,map[i][j].lat_inh_wts,lat_inh_dimension,inh_rad,i,j,
	  title,zaxis);
    break;
  }
  compute_mexhat(i,j);

  switch(output_routine){ /* Output the mexican hat stored in lat_inh_wts */

  case MATH: 
    mathematica(mexfile,Temp.lat_inh_wts,lat_inh_dimension,inh_rad,i,j);
    break;
  case MAPLE: 
    sprintf(title,"SUM OF LATERAL WTS OF A NEURON AT (%d,%d) IN A %dx%d GRID",i,j,N,N);
    sprintf(zaxis,"Exc+Inh");
    maple(mexfile,Temp.lat_inh_wts,lat_inh_dimension,inh_rad,i,j,
	  title,zaxis);
    break;
  }
}
/* ---------------------------------------------------------------------- */
void compute_mexhat(i,j) 
/* Compute hat for neuron (i,j). The exc and inh wts have already been scaled
 * by gammaexc*5 and gammainh*5. The mexican hat is returned in the inh wts.
 */
{
  int k, l, elowk, elowl, ehighk, ehighl,
      ilowk, ilowl, ihighk, ihighl;
  int exc_ar_width, inh_ar_width;
  double Wi;

  exc_ar_width = 2*exc_rad+1;
  inh_ar_width = 2*inh_rad+1;
  elowk  = i-exc_rad; if (elowk < 0) elowk =0    ;
  ehighk = i+exc_rad; if (ehighk>=N) ehighk = N-1;
  ilowk  = i-inh_rad; if (ilowk < 0) ilowk =0    ;
  ihighk = i+inh_rad; if (ihighk>=N) ihighk = N-1;
  elowl  = j-exc_rad; if (elowl < 0) elowl =0    ;
  ehighl = j+exc_rad; if (ehighl>=N) ehighl = N-1;
  ilowl  = j-inh_rad; if (ilowl < 0) ilowl =0    ;
  ihighl = j+inh_rad; if (ihighl>=N) ihighl = N-1;

  for(k=0; k< lat_inh_dimension; k++)       /* Copy the inhibition wts */
    Temp.lat_inh_wts[k]=map[i][j].lat_inh_wts[k];

  for(k=ilowk; k<=ihighk; k++)   /* Go through larger inhibitory surround */ 
    if ((k >=elowk) && (k <= ehighk)) /* For k values in excitatory range */
      for(l=ilowl; l<=ihighl; l++) 
	if ((l >= elowl) && (l <= ehighl)){ /* For l values in exc range  */
	  Wi= map[i][j].lat_exc_wts[(k-i+exc_rad)*exc_ar_width + (l-j+exc_rad)]  ;
	  Wi +=map[i][j].lat_inh_wts[(k-i+inh_rad)*inh_ar_width + (l-j+inh_rad)];
	  Temp.lat_inh_wts[(k-i+inh_rad)*inh_ar_width + (l-j+inh_rad)] = Wi ;
	}

}
/* ---------------------------------------------------------------------- */
void output(starti,startj)
int starti, startj; /* The starting neurons */
{
  int k,l;

  /* Read in snapshots */
  while ((fscanf(wtfile,"%d",&t)!=EOF) && (t <= snapshot))
    {
      weights_IO(fread,wtfile)    ;
    }

  switch(scanline){
  case HORIZ:               /* Horizontal Scanline */
    k=starti;
    for(l=startj; l < (startj+step*no_of_neurons); l+=step)
      dumpij(k,l,lookup_snaps[snap_index].offset,neuron_offset);

    dumppeaks(peakfile);
    dumpsigmoid(deltafile,betafile);
/*    dump_zwts(zfile); */
    break;

  case VERT:               /* Vertical Scanline   */
    l=startj;
    for(k=starti; k < (starti+step*no_of_neurons); k+=step)
      dumpij(k,l,lookup_snaps[snap_index].offset,neuron_offset);

    dumppeaks(peakfile);
    dumpsigmoid(deltafile,betafile);
 /*    dump_zwts(zfile); */
    break;
  }
  fclose(excfile); fclose(inhfile); fclose(mexfile); fclose(peakfile);
  fclose(deltafile); fclose(betafile);fclose(wtfile);
}
/* ---------------------------------------------------------------------- */
void maple(fp,wts,size,radius,ui,uj,title,zaxis)
FILE *fp;
double *wts;
int size,ui,uj;
char *title, *zaxis;
{
  int k, l, lowk, lowl, highk, highl;
  int ar_width;
  double Wi;

  ar_width = 2*radius+1;

  lowk  = ui-radius; if (lowk < 0) lowk =0    ;
  highk = ui+radius; if (highk>=N) highk = N-1;
  lowl  = uj-radius; if (lowl < 0) lowl =0    ;
  highl = uj+radius; if (highl>=N) highl = N-1;

 /* Print header for maple */
  fprintf(fp,"with(linalg);\nwith(plots);\n");
  fprintf(fp,"z := matrix( [ \n");

  for(k=0; k<lowk; k++){  /* For all  the zero connections */
    fprintf(fp, "[");
    for(l=0; l<N-1; l++)
      fprintf(fp,"0.0,");
    fprintf(fp,"0.0],\n");
  }
      
  for(k=lowk; k<=highk; k++){
    fprintf(fp, "[");

    for(l=0; l<lowl; l++)
      fprintf(fp,"0.0,");

    for(l=lowl; l<highl; l++) 
	fprintf(fp,"%lf,", 
		wts[(k-ui+radius)*ar_width + (l-uj+radius)]);
      
    fprintf(fp,"%lf",
	    wts[(k-ui+radius)*ar_width + (highl-uj+radius)]);

    for (l=highl+1; l < N; l++)
      fprintf(fp,",0.0");

    fprintf(fp,"]");
 
    if ( (k <= highk) && (k < N-1) ) fprintf(fp,",\n");
    else if (k == N-1) fprintf(fp," ] );\n");
  }

  for(k=highk+1; k<N; k++){  /* For all  the zero connections */
    fprintf(fp, "[");
    for(l=0; l<N-1; l++)
      fprintf(fp,"0.0,");
    
    if (k < N-1) fprintf(fp,"0.0],\n");
    else fprintf(fp,"0.0] ] );\n");
  }

  fprintf(fp,"matrixplot(z,title=`%s`,labels=[`x`,`y`,`%s`],axes=BOXED,style=PATCH,shading=Z);\n",title,zaxis);
}
/* ---------------------------------------------------------------------- */
void mathematica(fp,wts,size,radius,ui,uj)
FILE *fp; double *wts; 
int size,radius,ui,uj;
{
  int k, l, lowk, lowl, highk, highl;
  int ar_width;
  double Wi;

  ar_width = 2*radius+1;

  lowk  = ui-radius; if (lowk < 0) lowk =0    ;
  highk = ui+radius; if (highk>=N) highk = N-1;
  lowl  = uj-radius; if (lowl < 0) lowl =0    ;
  highl = uj+radius; if (highl>=N) highl = N-1;

  for(k=0; k<lowk; k++){  /* For all  the zero connections */
    for(l=0; l<N-1; l++)
      fprintf(fp,"0.0 ");
    fprintf(fp,"0.0\n");
  }

  for(k=lowk; k<=highk; k++){   /* Go through larger inhibitory surround */
  
    for(l=0; l<lowl; l++)
      fprintf(fp,"0.0 ");

    for(l=lowl; l<highl; l++) 
      fprintf(fp,"%lf ", 
	      wts[(k-ui+radius)*ar_width + (l-uj+radius)]);

    fprintf(fp,"%lf",
	    wts[(k-ui+radius)*ar_width + (highl-uj+radius)]);

    for (l=highl+1; l < N; l++)
      fprintf(fp," 0.0");

    fprintf(fp,"\n");
  }
  for(k=highk+1; k<N; k++){  /* For all  the zero connections */
     for(l=0; l<N-1; l++)
       fprintf(fp,"0.0 ");
    
    fprintf(fp,"0.0\n");
   }
}
/* ---------------------------------------------------------------------- */
double findpeak()
{
  int k; double peak;

  for(peak=0.0, k=0; k<lat_inh_dimension; k++)
    if (peak < Temp.lat_inh_wts[k]) peak = Temp.lat_inh_wts[k];
  return(peak);
}

void dumppeaks(fp) /* The peaks are found for each mexhat by searching */
FILE *fp;
{
  int i,j,k;
  int exc_ar_width, inh_ar_width;
  double peak;

  exc_ar_width = 2*exc_rad+1;
  inh_ar_width = 2*inh_rad+1;

  switch(output_routine){
  case MATH:
    for(i=0; i<N; i++)
      {
	for(j=0; j<N-1; j++){
	  compute_mexhat(i,j);

	  peak = findpeak();
	  fprintf(fp,"%lf", peak);
	}
	compute_mexhat(i,N-1);
	peak = findpeak();
	fprintf(fp,"%lf\n",peak);
      }
    break;

  case MAPLE:
  /* Print header for maple */
    fprintf(fp,"with(linalg);\nwith(plots);\n");
    fprintf(fp,"z := matrix( [ \n");

    for(i=0; i<N; i++){
      fprintf(fp, "[");
      for(j=0; j<N-1; j++) 
	{
	  compute_mexhat(i,j);

	  peak = findpeak();
	  fprintf(fp,"%lf,", peak);
		
	}
      compute_mexhat(i,N-1);
      peak = findpeak();
      fprintf(fp,"%lf]",peak);
	 
      if (i==N-1) fprintf(fp," ] );\n");
      else fprintf(fp,",\n");
    }
    fprintf(fp,"matrixplot(z,title=`PEAKS OF MEXICAN HAT FOR EACH NEURON`,labels=[`x`,`y`,`Peak Exc+Inh`],axes=BOXED,style=PATCH,shading=Z);\n");
    break;
  }
}
/* ---------------------------------------------------------------------- */
/* This procedure does reading and writing of weights to a binary file. fp
 * is a binary file pointer.
 */
void weights_IO(dofun,fp)
size_t (*dofun)();
FILE *fp;
{
  int i,j;
  size_t nobj;

  for(i=0; i<N; i++)
    for(j=0; j<N; j++){
      if((nobj=(*dofun)((void*)(& map[i][j].delta), sizeof(double), 
			1, fp)) != 1)
	fprintf(stderr,"Error in IO - delta. nobj = %d\n",nobj);

      if((nobj=(*dofun)((void*)(& map[i][j].beta), sizeof(double), 
			1, fp)) != 1)
	fprintf(stderr,"Error in IO - beta. nobj = %d\n",nobj);

      if((nobj=(*dofun)((void*)map[i][j].weights, sizeof(double), 
			(input_dimension+1), fp)) != (input_dimension+1))
	fprintf(stderr,"Error in IO - input wts. nobj = %d\n",nobj);

      if((nobj=(*dofun)((void*)map[i][j].lat_exc_wts, sizeof(double), 
			lat_exc_dimension, fp)) != lat_exc_dimension)
	fprintf(stderr,"Error in IO-lat exc wts.\n");

      if((nobj=(*dofun)((void*)map[i][j].lat_inh_wts, sizeof(double), 
			lat_inh_dimension, fp)) != lat_inh_dimension)
	fprintf(stderr,"Error in IO - lat inh wts.\n");
    }
}
/* ---------------------------------------------------------------------- */
void free_all()
{
  int i, j;
 
  for(i=0; i<N; i++)
    for(j=0; j<N; j++){
      free( (void*)map[i][j].weights);
      free( (void*)map[i][j].lat_exc_wts);
      free( (void*)map[i][j].lat_inh_wts);
    }
  free((void*)Temp.lat_inh_wts);
}
/* ---------------------------------------------------------------------- */
void inverse_transform(weights, return_input, input_dimension)
double *weights, *return_input; int input_dimension;
{
/* This function inverts the given input_weights and maps the neuron's weights
 * back to the original angle space. Inverse of the function tocartesian
 * in lcsofm.c
 * The dimension to which it maps back depends on the variable input_dimension.
 */
  int k; double cosproduct;

  for (k=0,cosproduct=1; k<input_dimension; k++)
    {
      return_input[k] = asin( weights[k]/cosproduct );
      cosproduct *= cos(return_input[k]);
    }
}
/* ---------------------------------------------------------------------- */

void dump_zwts(zfile)
FILE *zfile;
{
  int i,j,k; double angle0, angle1, angle2;

  switch(output_routine){
  case MATH:
    for(i=0; i<N; i++)
      {
	for(j=0; j<N-1; j++){
	  
	  angle0 = asin( map[i][j].weights[0] );
	  angle1 = asin( map[i][j].weights[1] / cos(angle0) );
	  angle2 = asin( map[i][j].weights[2] / (cos(angle1) * cos(angle0)) );

	  fprintf(zfile,"%lf", angle2);
	}
	angle0 = asin( map[i][N-1].weights[0] );
	angle1 = asin( map[i][N-1].weights[1] / cos(angle0) );
	angle2 = asin( map[i][N-1].weights[2] / (cos(angle1) * cos(angle0)) );
	fprintf(zfile,"%lf\n",angle2);
      }
    break;

  case MAPLE:
  /* Print header for maple */
    fprintf(zfile,"with(linalg);\nwith(plots);\n");
    fprintf(zfile,"z := matrix( [ \n");

    for(i=0; i<N; i++){
      fprintf(zfile, "[");
      for(j=0; j<N-1; j++) {
	  angle0 = asin( map[i][j].weights[0] );
	  angle1 = asin( map[i][j].weights[1] / cos(angle0) );
	  angle2 = asin( map[i][j].weights[2] / (cos(angle1) * cos(angle0)) );
	  fprintf(zfile,"%lf,", angle2 );
	}
      angle0 = asin( map[i][N-1].weights[0] );
      angle1 = asin( map[i][N-1].weights[1] / cos(angle0) );
      angle2 = asin( map[i][N-1].weights[2] / (cos(angle1) * cos(angle0)) );
      fprintf(zfile,"%lf]",angle2);
	 
      if (i==N-1)
	fprintf(zfile," ] );\n");

      else
	fprintf(zfile,",\n");
    }
    fprintf(zfile,"matrixplot(z,title=`Z values)`,labels=[`x`,`y`,`d`],axes=BOXED,style=PATCH,shading=Z);\n");
    break;
  }
}

