/* ----------------------------- File :xdraw.c ----------------------------- *
 * XWindow drawing routines for visualizing the network.
 * Features:
 * 
 * ------------------------------------------------------------------------- */

#include "latdefs.h"          /* Includes Xlib.h, Xutil.h, Xm.h, 
                               * stdio.h and math.h                         */

#define MAXLINES   1024       /* To draw upto 1024 lines in a single call.  */
#define MAXPOINTS  1024       /* To draw upto 1024 points in a single call. */

extern double distr_weight(); /* lcsofm.c: Used to calculate ideal gridpts  */
extern double scale01()     ; /* inputs.c: Used to scale to 0-1 range       */
extern int Activity_Color_Range; /* Set by display_init()                   */
extern XColor colors[255]   ;

void display_input()        ;
void display_net()          ;
void draw_grid()            ;
void display_activity()     ;
void draw_border()          ;
void inverse_transform()    ;
void display_dimension()    ;
/* ------------------------------------------------------------------------ */
void display_input(d, input_list)          
/* This procedure takes the first two elements of the raw input vectors
 * in the input_list as x,y co-ordinates in a 2d space and plots them.
 *
 * The foreground color has already been set in display_init.
 */
My_Display *d;
Input_history *input_list;
{ 
  extern Pixel newinput_color; /* Set in display_init()  */
  int x, y, count            ;
  Input_history *ptr         ;
  XPoint Points[MAXPOINTS]   ;

  if (input_list != NULL){
   
    XClearWindow(d->input.d, d->input.w);

    x = (int)ROUND((double)d->input.width  * 
		   scale01(0,input_list->input_vector[0]))     ;
    y = (int)ROUND((double)d->input.height *
		   (1-scale01(1,input_list->input_vector[1])) );

    /* The latest point (at the head of the list )should appear bigger -
     *  so we draw a filled rectangle centered around the actual x,y point.
     */
    XSetForeground(d->input.d, d->input.gc, newinput_color)            ;
    XFillRectangle(d->input.d, d->input.w, d->input.gc, x-1, y-1, 3, 3);
    XSetForeground(d->input.d, d->input.gc, d->input.fg)               ;
    
    for(count=0, ptr=input_list->next; ptr!=NULL; ptr=ptr->next)
      {
	Points[count].x   = (int)ROUND((double)d->input.width   * 
				     scale01(0,ptr->input_vector[0]))    ;
	Points[count++].y = (int)ROUND((double)d->input.height  * 
				    (1-scale01(1,ptr->input_vector[1]))) ;
	
	if (count==MAXPOINTS){
	  XDrawPoints(d->input.d, d->input.w, d->input.gc, Points,
		      count, CoordModeOrigin)                          ;
	  count = 0                                                    ;
	}
      }
    if (count>0){
      XDrawPoints(d->input.d, d->input.w, d->input.gc, Points,
		  count, CoordModeOrigin)                              ;
      count = 0                                                        ;
    }
    draw_border(d->input.d, d->input.w, d->input.gc,
	      d->input.width, d->input.height, d->input.fg)            ;
    XFlush(d->input.d)                                                 ;
  }
}
/* ---------------------------------------------------------------------- */
void display_net(d, net, netsize, border) 
/* Assumes that the first 2 elements of the input vector are the (x,y)
 * values in a 2d plane. The procedure maps the first three weight vectors
 * to this 2d plane, and plots a grid. The drawing color is the foreground
 * color of the display.
 */
My_Display *d;
Neuron net[NMAX][NMAX];
int netsize;
int border; /* No of neurons in the border of the network */
{
  extern Pixel idealgrid_color    ;
  int i,j,width,height,count=0    ;
  double x0, y0                   ;
  XSegment lines[MAXLINES]        ;
  XPoint gridpoints[NMAX][NMAX]   ; /* The points of the 2d grid */

  XClearWindow(d->net.d, d->net.w);
  
  /* -------------------- draw the ideal grid ----------------------------- */ 

  for(i=border; i < netsize-border; i++)  
    for(j=border; j< netsize-border; j++) 
      {                                   
	x0 = (double)d->net.width  * 
	  scale01(0,distr_weight(0,i-border,(netsize-2*border),0.0));
	y0 = (double)d->net.height * 
	  (1 - scale01(1,distr_weight(1,j-border,(netsize-2*border),0.0) ) );

	gridpoints[i-border][j-border].x = (int)ROUND(x0);
	gridpoints[i-border][j-border].y = (int)ROUND(y0);
      }
  XSetForeground(d->net.d, d->net.gc, idealgrid_color)   ;
  draw_grid(d->net.d,d->net.w,d->net.gc,gridpoints,netsize-2*border,lines);
  XSetForeground(d->net.d, d->net.gc, d->net.fg);/* Reset foreground color  */

  /* ---------------- draw the actual wt vector grid ---------------------- */ 

  for(i=border; i < netsize-border; i++)
    for(j=border; j< netsize-border; j++)
      {
	x0 = (double)d->net.height * 
	     (1 - scale01(0,asin(net[i][j].weights[0]) ) );
	
	y0 = (double)d->net.width  * 
	  scale01(1,asin(
		       net[i][j].weights[1] / cos( asin(net[i][j].weights[0]) )
		       ));

	gridpoints[i-border][j-border].x = (int)ROUND(x0);
	gridpoints[i-border][j-border].y = (int)ROUND(y0);
      }

  draw_grid(d->net.d,d->net.w,d->net.gc,gridpoints,netsize-2*border,lines);

  /* Draw a square to show the border of the network */
  draw_border(d->net.d, d->net.w, d->net.gc,
	      d->net.width, d->net.height, d->net.fg);

  XFlush(d->net.d);
}
	
/* ------------------------------------------------------------------------ */

void draw_grid(d, w, gc, gridpoints, gridsize, lines) 
Display *d; Window w; GC gc;XPoint gridpoints[NMAX][NMAX];
int gridsize ; XSegment *lines;
{
  int i, j, count=0;
  
  for(i=0; i < gridsize; i++)
    for(j=0; j < gridsize; j++)
      {
	if (i < gridsize-1) /* Draw horizontal lines upto gridsize-1 */
	  {
	    lines[count].x1   = gridpoints[i][j].x  ;
	    lines[count].y1   = gridpoints[i][j].y  ;
	    lines[count].x2   = gridpoints[i+1][j].x;
	    lines[count++].y2 = gridpoints[i+1][j].y;

	    if (count == MAXLINES-1){
		XDrawSegments(d, w, gc, lines, MAXLINES);
		count = 0;
	      }
	  }
	if (j < gridsize-1)  /* Draw vertical lines upto gridsize-1  */
	  {
	    lines[count].x1   = gridpoints[i][j].x  ;
	    lines[count].y1   = gridpoints[i][j].y  ;
	    lines[count].x2   = gridpoints[i][j+1].x;
	    lines[count++].y2 = gridpoints[i][j+1].y;

	    if (count == MAXLINES-1){
		XDrawSegments(d, w, gc, lines, MAXLINES);
		count = 0;
	      }
	  }
      }
  if (count > 0){                              /* Draw the remaining */
    XDrawSegments(d, w, gc, lines, count);
    count = 0;
  }
}
	
/* ---------------------------------------------------------------------- */

void display_activity(d, map_activity, netsize, maxact_i, maxact_j)
My_Display *d         ;
double map_activity[NMAX][NMAX];
int netsize           ;
int maxact_i, maxact_j;
{ 
  extern Pixel newinput_color; /* Set in display_init()  */
  int width, height, color_index, i, j;
  int x0,y0,x1,y1     ;                 /* The corners of each small square  */
  double maxsqx,maxsqy;
  int sqx, sqy        ;                 /* The size of the small rectangles  */
  double accumulate_x, accumulate_y;    /* To add up the x and y             */
  int count=0         ;
  int markersize      ;                 /* Marker for max.act neuron         */

  width  = d->net_act.width  ; 
  height = d->net_act.height ;
  
  /* Leave two pixels on each side for drawing the border lines */
  maxsqx = ((double)width-2.0)/(double)netsize ; 
  maxsqy = ((double)height-2.0)/(double)netsize;
  sqx    = (int)ROUND(maxsqx); sqy = (int)ROUND(maxsqy);
  markersize = (int)ROUND(maxsqx/4.0+1.0);

  XClearWindow(d->net_act.d, d->net_act.w);

  x0 = 1; accumulate_x = 1.0; /* Starting pixel : pixel 1 (0 for border)  */
  for( i=0; i < netsize; i++)
    {
      y0 = 1;accumulate_y=1.0      ;
      accumulate_x += maxsqx       ;
      x1 = (int) ceil(accumulate_x);
      for(j= 0; j < netsize; j++)
	{
	  color_index = (int)ROUND( (map_activity[i][j] *
				     (double)(Activity_Color_Range)) );
	  accumulate_y += maxsqy       ;
	  y1 = (int) ceil(accumulate_y);

	  XSetForeground(d->net_act.d, d->net_act.gc, 
			 colors[color_index].pixel) ;
	  XFillRectangle(d->net_act.d, d->net_act.w, d->net_act.gc, 
			 x0, (height-y1), sqx, sqy) ;

	  if ( (i==maxact_i) && (j==maxact_j) ) /* Mark max_act neuron */
	    {
	      XFlush(d->net_act.d);
	      XSetForeground(d->net_act.d, d->net_act.gc, 
			     newinput_color);
	      XFillRectangle(d->net_act.d, d->net_act.w, d->net_act.gc, 
			     x0, (height-y1),
			     markersize,markersize);
	    }

	  y0 = y1;
	}
      x0 = x1;
    }

  draw_border(d->net_act.d, d->net_act.w, d->net_act.gc,
	      d->net_act.width,d->net_act.height,d->net_act.fg);
  XFlush(d->net_act.d);
}	    

/* ---------------------------------------------------------------------- */

void draw_border(d, w, gc, width, height, fg)
Display *d; Window w;GC gc;
int width,height; Pixel fg;
{
  XSegment lines[4]; int count = 0;

  XSetForeground(d, gc, fg);
  lines[count].x1 = 0      ; lines[count].y1   = 0       ;
  lines[count].x2 = width-1; lines[count++].y2 = 0       ;
  lines[count].x1 = width-1; lines[count].y1   = 0       ;
  lines[count].x2 = width-1; lines[count++].y2 = height-1;
  lines[count].x1 = width-1; lines[count].y1   = height-1;
  lines[count].x2 = 0      ; lines[count++].y2 = height-1;
  lines[count].x1 = 0      ; lines[count].y1   = height-1;
  lines[count].x2 = 0      ; lines[count++].y2 = 0       ;
  XDrawSegments(d, w, gc, lines, 4) ;
}
/* ---------------------------------------------------------------------- */
void inverse_transform(weights, return_wts, input_dimension)
double *weights, *return_wts; int input_dimension;
{
/* This function inverts the given input_weights and maps the neuron's weights
 * back to the original angle space. Inverse of the function tocartesian
 * in lcsofm.c
 * The dimension to which it maps back depends on the variable input_dimension.
 */
  int k; double cosproduct;

  for (k=0, cosproduct=1; k<input_dimension; k++)
    {
      return_wts[k] = asin( weights[k]/cosproduct );
      cosproduct *= cos(return_wts[k]);
    }
}
/* ---------------------------------------------------------------------- */
void display_dimension(d,net,netsize,border,which_dimension)
My_Display *d; Neuron net[NMAX][NMAX]; 
int netsize, border, which_dimension;
{ 
/* To visualize how the neuron represents the specified dimension over the
 * network. Remember that the count for the which_dimension parameter starts
 * from zero. So if you wanted to see the third dimension, the which_dimension
 * value should be 2.
 * The colors used are the same as that for displaying activity. So white
 * indicates a maximum value for the dimension and black shows minimum value.
 */
  extern Pixel newinput_color;          /* Set in display_init()             */
  extern int input_dimension ;          /* declared in globals.h             */
  int width, height, color_index, i, j;
  int x0,y0,x1,y1     ;                 /* The corners of each small square  */
  double maxsqx,maxsqy;
  int sqx, sqy        ;                 /* The size of the small rectangles  */
  double accumulate_x, accumulate_y;    /* To add up the x and y             */
  double angle0, angle1, angle2    ;    /* To calculate corresponding dim.   */
  int count=0         ;
  double map_wts_back[MAXDIM-1]    ; /* Holds wts mapped back to input space */

  width  = d->att_act.width  ; 
  height = d->att_act.height ;
  
  /* Leave two pixels on each side for drawing the border lines */
  maxsqx = ((double)width-2.0)/(double)netsize ; 
  maxsqy = ((double)height-2.0)/(double)netsize;
  sqx    = (int)ROUND(maxsqx); sqy = (int)ROUND(maxsqy);

  XClearWindow(d->att_act.d, d->att_act.w);

  x0 = 1; accumulate_x = 1.0; /* Starting pixel : pixel 1 (0 for border)  */
  for( i=0; i < netsize; i++)
    {
      y0 = 1;accumulate_y=1.0      ;
      accumulate_x += maxsqx       ;
      x1 = (int) ceil(accumulate_x);
      for(j= 0; j < netsize; j++)
	{
	  if (((i < border) && (j < border)) || 
	      ((i>=netsize-border) && (j>=netsize-border))
	      )
	    color_index = newinput_color;
	  else{
	    inverse_transform(net[i][j].weights,map_wts_back,input_dimension);

#ifdef DEBUG3    
	    printf("(i,j,dim,z)=(%d,%d,%d,%.6lf): scale01(z) = %.6lf\n",
		   i,j,which_dimension,map_wts_back[which_dimension],
		   scale01(which_dimension, map_wts_back[which_dimension]));
#endif

	    color_index = (int)ROUND(scale01(which_dimension,
					     map_wts_back[which_dimension])
				     * (double)Activity_Color_Range);
	  }
	  accumulate_y += maxsqy       ;
	  y1 = (int) ceil(accumulate_y);

	  XSetForeground(d->att_act.d, d->att_act.gc, 
			 colors[color_index].pixel) ;
	  XFillRectangle(d->att_act.d, d->att_act.w, d->att_act.gc, 
			 x0, (height-y1), sqx, sqy) ;
	  y0 = y1;
	}
      x0 = x1;
    }

  draw_border(d->att_act.d, d->att_act.w, d->att_act.gc,
	      d->att_act.width,d->att_act.height,d->att_act.fg);
  XFlush(d->att_act.d);
}	    

/* ---------------------------------------------------------------------- */
